import json
import pytest
import os
import numpy as np
from skimage import io
from skimage.morphology import dilation
from unittest.mock import patch, MagicMock
from PIL import Image
from torchvision.transforms import ToPILImage
import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
    
from rhizonet.unet2D import ImageDataset, tiff_reader
from rhizonet.utils import extract_largest_component_bbox_image


@pytest.fixture
def mock_images_files(tmp_path):
    image_files = [tmp_path / f"image_{i}.tif" for i in range(5)]
    label_files = [tmp_path / f"label_{i}.png" for i in range(5)]
    for img, lab in zip(image_files, label_files):
        img.touch()
        lab.touch()
    return image_files, label_files


@pytest.fixture
def synthetic_sample(tmp_path):
    """
    Create synthetic image and label sample as temporary files for testing.
    """
    image_path = tmp_path / "test_image.tif"
    label_path = tmp_path / "test_label.tif"
    image_data = (np.random.rand(256, 256, 3) * 255).astype(np.uint8)
    label_data = (np.random.randint(0, 2, size=(256, 256)) * 255).astype(np.uint8)
    Image.fromarray(image_data).save(image_path)
    io.imsave(label_path, label_data)

    return {
        "image_path": str(image_path),
        "label_path": str(label_path),
        "image_data": image_data,
        "label_data": label_data,
    }


@pytest.fixture
def mock_args():
    return {
    "input_channels": 3,
    "class_values": (0, 85, 170),
    "data_split": [
      0.8,
      0.1,
      0.1
    ],
    "translate_range": 0.2,
    "rotate_range": 0.05,
    "scale_range": 0.1,
    "shear_range": 0.1,
    "patch_size": (
      64,
      64
    ),
    "image_col": "None",
    "boundingbox": True,
    "dilation": True,
    "disk_dilation": 2
    }


def test_mismatched_data_label(mock_images_files, mock_args):
    data_fnames, label_fnames = mock_images_files
    label_fnames = label_fnames[:-1]  # Remove one label to cause a mismatch
    with pytest.raises(SystemExit):
        ImageDataset(data_fnames, label_fnames, mock_args)


def test_tiff_reader(synthetic_sample):
    reader = tiff_reader(
        image_col=None,
        boundingbox=True,
        dilation=True,
        disk_dilation=3,
        keys=["image", "label"]
    )
    data_dict = {
        "image": synthetic_sample["image_path"],
        "label": synthetic_sample["label_path"],
    }

    transformed_data = reader(data_dict)
    assert "image" in transformed_data
    assert "label" in transformed_data

    # Check shapes and types
    assert transformed_data["image"].shape[0] == 3  # Channel dimension should be first
    assert isinstance(transformed_data["image"], np.ndarray)
    assert transformed_data["label"].shape[0] == 1  # Channel dimension should be first for labels
    assert isinstance(transformed_data["label"], np.ndarray)


def test_invalid_input_shape(tmp_path):

    topil = ToPILImage()
    invalid_image_path = tmp_path / "invalid_image.tif"

    invalid_image_data = (np.random.rand(256, 256) * 255).astype(np.float32)
    invalid_image_data = topil(invalid_image_data)
    invalid_image_data.save(invalid_image_path)

    invalid_label_data = np.random.randint(0, 2, (256,256)).astype(np.uint8)
    invalid_label_path =  tmp_path / "invalid_label.tif"

    io.imsave(invalid_label_path, invalid_label_data)


    reader = tiff_reader(
        image_col=None,
        boundingbox=True,
        dilation=True,
        disk_dilation=3,
        keys=["image", "label"]
    )

    data_dict = {
        "image": str(invalid_image_path),
        "label": str(invalid_label_path),  
    }

    # Expecting a ValueError for invalid input shape
    with pytest.raises(ValueError, match=r"Unexpected image shape"):
        reader(data_dict)


def test_dataset_initialization(mock_images_files, mock_args):

    data_fnames, label_fnames = mock_images_files
    dataset = ImageDataset(data_fnames, label_fnames, mock_args)

    assert dataset.Nsamples == len(data_fnames)
    assert dataset.patch_size == mock_args['patch_size']
    assert dataset.spatial_dims == len(mock_args['patch_size'])
    assert isinstance(dataset.boundingbox, bool)
    assert isinstance(dataset.dilation, bool) 


def test_transforms_training(mock_args):
    dataset = ImageDataset([], [], mock_args, training=True)
    transforms = dataset.get_data_transforms(training=True, boundingbox=None, dilation=False, disk_dilation=2)
    assert "RandFlipd" in str(transforms.transforms)
    assert "RandAffined" in str(transforms.transforms)
    assert "MapLabelValued" in str(transforms.transforms)
    assert "CastToTyped" in str(transforms.transforms)
    assert "EnsureTyped" in str(transforms.transforms)


def test_transforms_validation(mock_args):
    dataset = ImageDataset([], [], mock_args, training=False)
    transforms = dataset.get_data_transforms(training=False, boundingbox=None, dilation=1, disk_dilation=2)
    assert "RandFlipd" not in str(transforms.transforms)
    assert "RandAffined" not in str(transforms.transforms)
    assert "Resized" in str(transforms.transforms)
    assert "MapLabelValued" in str(transforms.transforms)
    assert "CastToTyped" in str(transforms.transforms)
    assert "EnsureTyped" in str(transforms.transforms)


if __name__ == "__main__":
    pytest.main(["-v", __file__])




import pytest
import numpy as np
import os
from argparse import Namespace
import json
from skimage import io, util
from tempfile import TemporaryDirectory
import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
from rhizonet.prepare_patches import parse_prepare_variables, prepare_patches

@pytest.fixture
def sample_config():
    return {
        "images_path": "./images",
        "labels_path": "./labels",
        "save_path": "./output",
        "train_patch_size": [64, 64],
        "step_size": [32, 32],
        "min_labeled": 0.1,
        "nb_pred_data": None,
        "label_prefix": ""
    }

def test_parse_prepare_variables(sample_config):
    sample_config = {
        "images_path": "./images",
        "labels_path": "./labels",
        "save_path": "./output",
        "train_patch_size": [64, 64],
        "step_size": [32, 32],
        "min_labeled": 0.1,
        "nb_pred_data": None,
        "label_prefix": ""
    }

    with TemporaryDirectory() as temp_dir:
        config_path = os.path.join(temp_dir, "config.json")
        with open(config_path, 'w') as f:
            json.dump(sample_config, f)

        args = Namespace(config_file=config_path)
        parsed_args = parse_prepare_variables(args)

        assert parsed_args['train_patch_size'] == (64, 64)
        assert parsed_args['step_size'] == (32, 32)
        assert parsed_args['min_labeled'] == 0.1

def test_prepare_patches():
    sample_config = {
        "images_path": "./images",
        "labels_path": "./labels",
        "save_path": "./output",
        "train_patch_size": (64, 64),
        "step_size": (32, 32),
        "min_labeled": 0.1,
        "nb_pred_data": None,
        "label_prefix": "Annotation"
    }
    with TemporaryDirectory() as temp_dir:
        images_dir = os.path.join(temp_dir, "images")
        labels_dir = os.path.join(temp_dir, "labels")
        os.makedirs(images_dir)
        os.makedirs(labels_dir)

        for i in range(2):
            img = np.random.randint(0, 256, (128, 128, 3), dtype=np.uint8)
            label = np.zeros((128, 128), dtype=np.uint8)
            label[32:96, 32:96] = 1  # Annotated region

            io.imsave(os.path.join(images_dir, f"image_{i}.png"), img)
            io.imsave(os.path.join(labels_dir, f"Annotationimage_{i}.png"), label)

        sample_config["images_path"] = images_dir
        sample_config["labels_path"] = labels_dir
        sample_config["save_path"] = os.path.join(temp_dir, "output")

        prepare_patches(sample_config)

        output_images_dir = os.path.join(sample_config['save_path'], "images")
        output_labels_dir = os.path.join(sample_config['save_path'], "labels")

        assert os.path.exists(output_images_dir)
        assert os.path.exists(output_labels_dir)

        output_images = sorted(os.listdir(output_images_dir))
        output_labels = sorted(os.listdir(output_labels_dir))

        assert len(output_images) > 0
        assert len(output_images) == len(output_labels)

        for label_file in output_labels:
            label = io.imread(os.path.join(output_labels_dir, label_file))
            assert (np.count_nonzero(label) / label.size) > sample_config['min_labeled']


if __name__ == "__main__":
    pytest.main(["-v", __file__])


"""
Script for evaluation metrics on a given set of predictions and groundtruth. 
The metrics used are Accuracy, Precision, Recall and IOU. 

Usage:
    pip install rhizonet
    evalmetrics_rhizonet ---pred_path "path" --label_path "path" --log_dir "path" --task "binary" --num_classes "2" --frg_class 85
"""

import os
import torch
import json 
import numpy as np
from argparse import ArgumentParser
import torchmetrics
from skimage import io
from typing import Tuple, Sequence, List

try:
    from .utils import createBinaryAnnotation, MapImage
except ImportError:
    from utils import createBinaryAnnotation, MapImage


def calculate_all_metrics(pred: torch.Tensor, 
                         target: torch.Tensor,  
                         task: str = 'binary', 
                         num_classes: int = 2, 
                         background_index: int = 0) -> Tuple[float, float, float, float]:
    """
    Evaluate Accuracy, Precision, Recall and IOU based on prediction and groundtruth and for a given ask (e.g. `binary`).

    Args:
        pred (torch.Tensor): predicted image
        target (torch.Tensor): groundtruth image
        task (str, optional): type of classification for each pixel (binary or multi-class). Defaults to 'binary'.
        num_classes (int, optional): number of classes. Defaults to 3.
        background_index (int, optional): index value associated to the background. Defaults to 0.

    Returns:
        Tuple[float, float, float, float]: averaged accuracy, precision, recall and IOU
    """

    cnfmat = torchmetrics.ConfusionMatrix(
                                        num_classes=num_classes,
                                        task=task,
                                        normalize=None
                                        )

    cnfmat = cnfmat(pred, target)
    true = torch.diag(cnfmat)
    tn = true[background_index]
    tp = torch.cat([true[:background_index], true[background_index + 1:]])

    fn = (cnfmat.sum(1) - true)[torch.arange(cnfmat.size(0)) != background_index]
    fp = (cnfmat.sum(0) - true)[torch.arange(cnfmat.size(1)) != background_index]

    acc = torch.sum(true) / torch.sum(cnfmat)
    precision = torch.sum(tp) / torch.sum(tp + fp)
    recall = torch.sum(tp) / torch.sum(tp + fn)
    iou = torch.sum(tp) / (torch.sum(cnfmat) - tn)
    iou_per_class = tp / (tp + fp + fn)
    dice = 2 * torch.sum(tp) / (2 * torch.sum(tp) + torch.sum(fp) + torch.sum(fn))
    
    return acc.item(), precision.item(), recall.item(), iou.item(), dice.item()


def evaluate(pred_path: str, label_path: str, log_dir: str, task: str, num_classes: int, frg_class: int = 255) -> None:
    """
    Reads the prediction and groundtruth images, evaluates the metrics Accuracy, Precision, Recall and IOU.
    Saves results in `metrics.json` file in the specified `log_dir`. 
    There are 2 options: 
    - evaluate metrics on multi-class prediction.
    - evaluate metrics on binary segmentation mask (e.g. if the prediction is processed into a binary mask with root as foreground and the rest labeled as background.)

    Args:
        pred_path (str): filepath of the predicted image
        label_path (str): filepath of the groundtruth image
        log_dir (str): filepath where results will be saved in a json file
        num_classes (int): number of class labels
        task (int): type of segmentation task if processing binary segmentation masks or multiclass segmentation images (e.g. `binary` or `multiclass`)
        frg_class (int): value of the foreground class when creating binary segmentation masks

    """
    pred_list = sorted([os.path.join(pred_path, e) for e in os.listdir(pred_path) if not e.startswith(".")])
    label_list = sorted([os.path.join(label_path, e) for e in os.listdir(label_path) if not e.startswith(".")])
    dict_metrics = {}


    # if number of classes in the prediction is 2 then binary else multiclass with number of classes 
    for pred_path, lab_path in zip(pred_list, label_list):
        pred = io.imread(pred_path) # Predicted image 
        lab = io.imread(lab_path) # Groundtruth image

        if task == 'binary':
            lab = createBinaryAnnotation(lab, frg_class)
            lab = torch.Tensor(lab/255.0)

        # Check if prediction is scaled by 255 for visualization 
        if np.min(pred) >= 0 and np.max(pred) == 255:
            pred = torch.Tensor(pred/255.0)

        original_labels = np.unique(pred)
        pred = MapImage(torch.tensor(pred), original_labels, reverse=False)
        lab = MapImage(torch.tensor(lab), original_labels, reverse=False)
        filename = os.path.basename(pred_path)

        dict_metrics[filename] = {}
        acc, prec, rec, iou, dice = calculate_all_metrics(pred, lab, task=task, num_classes=num_classes)
        dict_metrics[filename]['accuracy'] = acc
        dict_metrics[filename]['precision'] = prec
        dict_metrics[filename]['recall'] = rec
        dict_metrics[filename]['IOU'] = iou
        dict_metrics[filename]['Dice'] = dice
                                                                                                                                                 
    # print("Metrics: \n {}".format(dict_metrics))
    with open(os.path.join(log_dir, 'metrics.json'), 'w') as f:
        json.dump(dict_metrics, f)
        
def main():
    parser = ArgumentParser(conflict_handler='resolve', description='Run inference using trained model')
    parser.add_argument("--pred_path", type=str,
                        default=".results/training_patches64_ex7ex9_batch32_dropout40/predictions/",
                        help="path to the predictions")
    parser.add_argument("--label_path", 
                        default="path to the groundtruth")
    parser.add_argument("--log_dir",
                        default="/home/zsordo/rhizonet-fovea/results/training_patches64_ex7ex9_batch32_dropout40",
                        help="path where the json file containing results will be saved")
    
    parser.add_argument("--task",
                        default='binary',
                        choices=['binary', 'multiclass'],
                        help="type of segmentation task if processing binary segmentation masks or multiclass segmentation images")
    parser.add_argument("--num_classes",
                        default=2,
                        help="Number of classes including background")
    parser.add_argument("--frg_class",
                        default=255,
                        help="value of the foreground class when using binary segmentation masks")
    
    args = parser.parse_args()
    args = vars(args)
    evaluate(args['pred_path'], args ['label_path'], args['log_dir'], args['task'], args['num_classes'], args['frg_class'])

if __name__ == '__main__':
    main()


"""
Script for running inference using a pre-trained residual U-Net model for image segmentation.

Usage:
    pip install rhizonet
    predict_rhizonet --config_file ./setup_files/setup-predict.json 
"""

import os
import glob
import argparse
import json
import re
from tqdm import tqdm
from argparse import ArgumentParser
from skimage import io, util, color, filters, exposure, measure
from PIL import Image
import numpy as np
import pytorch_lightning as pl
import torch
import torchmetrics
import matplotlib.pyplot as plt
from scipy import ndimage as ndi
from skimage.color import rgb2lab

from monai.data import ArrayDataset, create_test_image_2d, list_data_collate, decollate_batch
from monai.inferers import sliding_window_inference
from monai.transforms import (
    Compose,
    ScaleIntensityRange,
    EnsureType
)

from PIL import ImageDraw
import torchvision.transforms.functional as TF
from datetime import datetime

from typing import Tuple, List, Dict, Sequence, Union, Any
from collections.abc import Callable

# Import parent modules 
try:
    from .utils import MapImage, createBinaryAnnotation, extract_largest_component_bbox_image
    from .unet2D import Unet2D, PredDataset2D, ImageDataset, tiff_reader, dynamic_scale
except ImportError:
    from utils import MapImage, createBinaryAnnotation, extract_largest_component_bbox_image
    from unet2D import Unet2D, PredDataset2D, ImageDataset, tiff_reader, dynamic_scale


def _parse_training_variables(argparse_args) -> Dict:
    """ 
    Parse and merge training variables from a JSON configuration file and command-line arguments.

    Args:
        argparse_args (Namespace): Command-line arguments parsed by argparse.

    Returns:
        Dict: Updated arguments 
    """


    args = vars(argparse_args)
    # overwrite argparse defaults with config file
    with open(args["config_file"]) as file_json:
        config_dict = json.load(file_json)
        args.update(config_dict)
    args['pred_patch_size'] = tuple(args['pred_patch_size']) # tuple expected, not list
    if args['gpus'] is None:
        args['gpus'] = -1 if torch.cuda.is_available() else 0

    return args


def transform_image(img_path:str) -> Tuple[np.ndarray, str]:
    """
    Reads the filepath and returns the image in the correct shape for inference (C, H, W)

    Args:
        img_path (str): Filepath of the input image

    Returns:
        Tuple[np.ndarray, str]: Image in the correct shape, Filepath of the image 
    """
    transform = Compose(
        [
            EnsureType()
        ]
    )
    img = io.imread(img_path)# only 3 modalities are accepted in the channel dimension for now
    if img.ndim == 4 and img.shape[-1] < 4:  # If shape is (h, w, d, c) assuming there are maximum 4 channels or modalities 
        img = np.transpose(img[..., :3] , (3, 0, 1, 2))  # Move channel to the first position
        img = dynamic_scale(img)
    elif img.ndim == 3 and img.shape[-1] <= 4:  # If shape is (h, w, c)
        img = np.transpose(img[..., :3] , (2, 0, 1))  # Move channel to the first position
        img = dynamic_scale(img)
    elif img.ndim == 2: # if no batch dimension then create one
        img = np.expand_dims(img, axis=-1)
        img = dynamic_scale(img)
        img = np.transpose(img, (2, 0, 1))
    else:
        raise ValueError(f"Unexpected image shape: {img.shape}, channel dimension should be last and image should be either 2D or 3D")
        
    img = transform(img)
    return img, img_path


def pred_function(
        image : torch.Tensor, 
        model: Callable,
        pred_patch_size: Sequence[int]
        ) -> torch.Tensor:
    """
    Sliding window inference on `image` with `model`

    Args:
        image (torch.Tensor): input image to be processed
        model (Callable): given input tensor ``image`` in shape BCHW[D], the outputs of the function call ``model(input)`` should be a tensor.
        pred_patch_size (Sequence[int]): spatial window size for inference

    Returns:
        torch.Tensor: prediction tensor
    """
    
    return sliding_window_inference(inputs=image, roi_size=pred_patch_size, sw_batch_size=1, predictor=model)


def predict_step(
        image_path: str, 
        model: Callable,
        pred_patch_size: Sequence[int]
        ) -> torch.Tensor:
    """
    Call trained model and run inference on input image given by the filepath using monai's ``sliding_window_inference`` function. 

    Args:
        image_path (str): filepath of the input image to be processed
        model (Callable): given input tensor ``image`` in shape BCHW[D], the outputs of the function call ``model(input)`` should be a tensor.
        pred_patch_size (Sequence[int]): spatial window size for inference

    Returns:
        torch.Tensor: prediction obtained by:
            - using argmax (computes maximum value along the class dimension)   
            - casting the tensor to torch.uint8 (byte) and scaling to 255 for visualization
    """
    image, _ = transform_image(image_path)
    cropped_image = extract_largest_component_bbox_image(image.unsqueeze(0), lab=None)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tensor_cropped_image = torch.tensor(cropped_image).to(device)
    logits = pred_function(tensor_cropped_image.float(), model, pred_patch_size)
    pred = torch.argmax(logits, dim=1).byte().squeeze(dim=1)
    pred = (pred * 255).byte()
    return pred


def get_prediction(
        file: str, 
        unet: Callable,
        pred_patch_size: Sequence[int], 
        save_path: str, 
        labels: Sequence[int],
        binary_preds: bool,
        frg_class: int):
    """
    Convert the prediction to a binary segmentation mask and saves the image in the ``save_path`` filepath specified in the configuration file.

    Args:
        file (str): input image filepath
        unet (Callable): trained callable model
        pred_patch_size (Sequence[int]): spatial window size for inference
        save_path (str): path in which the predictions will be saved
        labels (Sequence[int]): the labels used for annotating the groundtruth
        binary_preds (bool): generate binary predictions (e.g. root vs background) or keep all class labels
        frg_class (int): value of the foreground class when using binary segmentation masks
    """

    prediction = predict_step(file, unet, pred_patch_size).squeeze(0)
    prediction = MapImage(prediction, labels, reverse=True)
    pred = prediction.cpu().numpy().squeeze().astype(np.uint8)
    # pred_img, mask = elliptical_crop(pred, 1000, 1500, width=1400, height=2240)
    if binary_preds:
        binary_mask = createBinaryAnnotation(pred, frg_class=frg_class).astype(np.uint8)
        io.imsave(os.path.join(save_path, os.path.basename(file).split('.')[0] + ".png"), binary_mask, check_contrast=False)
    else:
        io.imsave(os.path.join(save_path, os.path.basename(file).split('.')[0] + ".png"), pred.astype(np.unint8), check_contrast=False)


def predict_model(args: Dict):
    """
    Compile all functions above to run inference on a list of images

    Args:
        args (Dict): arguments specified in the configuration file
    """

    pred_data_dir = args['pred_data_dir']
    save_path = args['save_path']
    labels = args['labels']
    binary_preds = args['binary_preds']
    frg_class = args['frg_class']
    
    os.makedirs(save_path, exist_ok=True)
    
    # Looping through all ecofab folders in the pred_data_dir directory
    for ecofab in sorted(os.listdir(pred_data_dir)):
        if not ecofab.startswith("."):

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            unet = Unet2D.load_from_checkpoint(args['model_path']).to(device)
            unet.eval()

            if os.path.isdir(os.path.join(pred_data_dir, ecofab)): #if you have a timeseries of one ecofab 
                os.makedirs(os.path.join(save_path, ecofab), exist_ok=True)
                lst_files = sorted(os.listdir(os.path.join(pred_data_dir, ecofab)))
                for file in tqdm(lst_files):
                    if not file.startswith("."):
                        print("Predicting for {}".format(file))
                        file_path = os.path.join(pred_data_dir, ecofab, file)
                        get_prediction(file_path, unet, args['pred_patch_size'], os.path.join(save_path, ecofab), labels, binary_preds, frg_class)
            else:
                print("Predicting for {}".format(ecofab))
                file_path = os.path.join(pred_data_dir, ecofab)
                get_prediction(file_path, unet, args['pred_patch_size'], save_path, labels, binary_preds, frg_class) 

def main():

    parser = argparse.ArgumentParser(conflict_handler='resolve', description='Run inference using trained model')
    parser.add_argument("--config_file", type=str,
                        default="setup_files/setup-predict.json",
                        help="json file training data parameters")
    parser.add_argument("--gpus", type=int, default=None, help="how many gpus to use")
    args = parser.parse_args()
    args = _parse_training_variables(args)

    predict_model(args)

if __name__ == '__main__':
    main()



"""
Script for applying postprocessing methods to inferred predictions using a the convex hull method as well as morphological operations. 

Usage:
    postprocess_rhizonet --config_file ./setup_files/setup-processing.json 
"""

import os
import numpy as np
from skimage import io, util, color
import argparse
from tqdm import tqdm
import json
import torch
from typing import List, Union, Sequence, Tuple

from skimage.morphology import (erosion, dilation, closing, opening,
                                area_closing, area_opening)
from skimage.measure import label
from skimage.morphology import convex_hull_image, disk

def _parse_training_variables(argparse_args):
    """ 
    Parse processing variables from the JSON postprocessing configuration file and command-line arguments.

    Args:
        argparse_args (Namespace): Command-line arguments parsed by argparse.

    Returns:
        Dict: Updated arguments 
    """    

    args = vars(argparse_args)
    # overwrite argparse defaults with config file
    with open(args["config_file"]) as file_json:
        config_dict = json.load(file_json)
        args.update(config_dict)
    if args['gpus'] is None:
        args['gpus'] = -1 if torch.cuda.is_available() else 0

    return args

def getLargestCC(segments: np.ndarray) -> np.ndarray:
    """
    Find the largest connected component within a 2D image 

    Args:
        segments (np.ndarray): A 2D binary or multi-class array (NumPy array) where labeled regions are to be identified.
            - Binary image: Where non-zero values represent objects (foreground), and zero values represent the background.
            - Multi-class image: Where different integer values represent distinct classes.

    Returns:
        np.ndarray: A 2D  mask corresponding to the largest object.
    """
    labels = label(segments)
    largestCC = labels == np.argmax(np.bincount(labels.flat, weights=segments.flat))
    return largestCC


def maxProjection(limg: List[np.ndarray], ndown: int =1) -> np.ndarray:
    """
    Apply maximum projection to a list of 2D slices.
    The images in the list limg are downsampled by ndown before the maximum intensity projection is computed

    Args:
        limg (List[np.ndarray]): list of 2D images. 
        ndown (int, optional): integer parameter that controls the downsampling factor for the 2D images along dimensions x and y when performing the maximum projection. Defaults to 1 so no downsampling.

    Returns:
        np.ndarray: 2D projected image. 
    """
    #max projection aka. z-max
    IM_MAX = limg[0][::ndown,::ndown]
    for n in np.arange(1, len(limg),ndown): #ndown here for low variation on Z
        IM_MAX = np.maximum(IM_MAX, (limg[n][::ndown,::ndown]))
    return IM_MAX


def processing(data_path: str, 
               output_path: str, 
               area_opening_param: int = 500, 
               area_closing_param: int = 200, 
               disk_radius: int = 4):
    """
    Perform post-processing on predictions for noise removal, using the convex hull method and morphological operations. 

    Args:
        data_path (str): Data directory
        output_path (str): Directory where processed predictions will be saved
        area_opening_param (int, optional): minimum area of objects to retain when (eroding, dilating). Defaults to 500.
        area_closing_param (int, optional): minimum area of objects to fill when (dilating, eroding). Defaults to 200.
        disk_radius (int, optional): radius of the disk object to use when applying morphological operations. Defaults to 4.

    Returns: None
    """
    element = disk(disk_radius)

    for e in tqdm(sorted([e for e in os.listdir(data_path) if not e.startswith(".")])):
        pred_data = os.path.join(data_path, e)
        pred_chull_dir = os.path.join(output_path, e)

        if os.path.isdir(pred_data):
            os.makedirs(pred_chull_dir, exist_ok=True)

            # Get hull convex shape of overlapped images for one ecofolder
            lst_img = []
            for file in sorted([e for e in os.listdir(pred_data) if not e.startswith(".")]):
                print("Processing {}".format(file))

                path = os.path.join(pred_data, file)
                img = io.imread(path)
                lst_img.append(img)
            proj_img = maxProjection(lst_img)
        else:
            print("Processing {}".format(e))

            os.makedirs(output_path, exist_ok=True)
            img = io.imread(pred_data)
            proj_img = maxProjection([img])
            
        # apply morphological dilation
        dilated_img = dilation(proj_img, element)

        lcomponent = getLargestCC(dilated_img)
        chull = convex_hull_image(lcomponent)

        if os.path.isdir(pred_data): #if you have a timeseries of one ecofab 

            for j, file in enumerate(sorted([e for e in os.listdir(pred_data) if not e.startswith(".")])):
                path = os.path.join(pred_data, file)
                img = io.imread(path)
    
                result = np.zeros_like(img)
                result[chull == 1] = img[chull == 1]
    
                # apply morphological operations (area opening on area closing)
                pred = area_opening(area_closing(result, area_closing_param), area_opening_param)
    
                # apply additional erosion to obtain thinner roots
                pred_eroded = erosion(pred, disk(2))
                io.imsave(os.path.join(pred_chull_dir, file), pred_eroded, check_contrast=False)
                
        else: # if you have one image for a given ecofab instead of a timeseries
            result = np.zeros_like(img)
            result[chull == 1] = img[chull == 1]

            # apply morphological operations (area opening on area closing)
            pred = area_opening(area_closing(result, area_closing_param), area_opening_param)
            # apply additional erosion to obtain thinner roots
            pred_eroded = erosion(pred, disk(2))
            io.imsave(pred_chull_dir, pred_eroded, check_contrast=False)


def main():
    parser = argparse.ArgumentParser(conflict_handler='resolve', description="Arguments for applying Convx Hull post processing to a set of binary masks")
    parser.add_argument("--config_file", type=str,
                        default="./setup_files/setup-processing.json",
                        help="json file training data parameters")
    parser.add_argument("--gpus", type=int, default=None, help="how many gpus to use")

    args = parser.parse_args()
    args = _parse_training_variables(args)

    processing(args['data_path'], args['output_path'], args['area_opening'], args['area_closing'],args['disk_radius'])


if __name__ == '__main__':
    main()


import os
import torch
import numpy as np 
import pytest
import unittest
import numpy as np
import torch
from PIL import Image

import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
from rhizonet.utils import (
    extract_largest_component_bbox_image,
    get_weights,
    MapImage,
    elliptical_crop,
    get_image_paths,
    contrast_img,
    createBinaryAnnotation,
    get_biomass
)

# def test_extract_largest_cc_bbox():
#     img = torch.rand((3, 128, 128), dtype=torch.float32)
#     output = extract_largest_component_bbox_image(img, predict=True)
#     assert isinstance(output,torch.tensor)
#     assert output.ndim == 3


@pytest.mark.parametrize("include_background, expected_length", [
    (True, 3),
    # (False, 2)
])
def test_get_weights(include_background, expected_length):
    labels = torch.tensor([[0, 1, 2], [2, 0, 1]])
    classes = [0, 1, 2]
    weights = get_weights(labels, classes=classes, device='cpu', include_background=include_background)
    assert len(weights) == expected_length
    assert all(w >= 0 for w in weights)


def test_MapImage():
    allowed_values = [0, 85, 170]
    img = np.random.choice(allowed_values, size=(10, 10))
    mapped_values = [0, 1, 2]
    mapped_img = MapImage(img, original_values=allowed_values, reverse=False)
    assert mapped_img.shape == img.shape
    assert np.all(np.isin(mapped_img, mapped_values))


def test_elliptical_crop():
    img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    cropped_image, _ = elliptical_crop(img, 50, 50, 40, 60)
    assert cropped_image.size == (100, 100)

def test_get_image_paths(tmp_path):
    temp_dir = tmp_path / "test_images"
    temp_dir.mkdir()
    for i in range(3):
        (temp_dir / f"image_{i}.jpg").write_text("")
    paths = get_image_paths(str(temp_dir))
    assert len(paths) == 3


def test_contrast_img():
    img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    contrasted = contrast_img(img)
    assert contrasted.shape == img.shape[:-1]
    assert np.max(contrasted) <= 1.0


@pytest.mark.parametrize("img_values, expected_fg_value", [
    ([0, 1, 255], 255),
    ([0, 1], 255)
])
def test_createBinaryAnnotation(img_values, expected_fg_value):
    img = np.random.choice(img_values, (100, 100), replace=True).astype(np.uint8)
    binary_mask = createBinaryAnnotation(img, expected_fg_value)
    assert binary_mask.shape == img.shape
    assert np.all(np.isin(binary_mask, [0, expected_fg_value]))


def test_get_biomass():
    binary_img = np.zeros((100, 100), dtype=np.uint8)
    binary_img[20:40, 30:50] = 1
    biomass = get_biomass(binary_img)
    assert biomass == 400  # Area of 20x20 pixels = 400


if __name__ == "__main__":
    pytest.main(["-v", __file__])


import pytest
import os 
import numpy as np 
import torch
import tempfile

from argparse import Namespace
from argparse import ArgumentParser
from unittest.mock import Mock, MagicMock
from unittest.mock import patch
import wandb
from pytorch_lightning.loggers import WandbLogger
import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
from rhizonet.train import train_model, _parse_training_variables
from rhizonet.unet2D import Unet2D
import pytorch_lightning as pl


def test_unet2d_initialization():
    train_ds = MagicMock()  # Mock training dataset
    val_ds = MagicMock()    # Mock validation dataset
    
    model_params = {
    "input_channels": 3,
    "class_values": [0, 85, 170],
    "spatial_dims": 2, 
    "pred_patch_size": (
      64,
      64
    ),
    "task": "multiclass",
    "background_index": 0,
    "num_classes": 3,
    "model": "resnet",
    "batch_size": 1,
    "lr": 3e-4,
    "num_workers": 32,
    "nb_epochs": 1
    }

    # Valid initialization
    model = Unet2D(train_ds, val_ds, **model_params)
    assert isinstance(model, Unet2D), "Model initialization failed"
    
    # Invalid model type
    with pytest.raises(AttributeError):
        Unet2D(train_ds, val_ds, model='invalid_model_type', 
               num_classes=3, 
               pred_patch_size=(64, 64), 
               spatial_dims=2,
               class_values=[0, 85, 170],
               task='multiclass')


def test_forward_pass():
    train_ds = MagicMock()
    val_ds = MagicMock()

    # Test for resnet 
    model = Unet2D(train_ds, val_ds, model='resnet', 
                input_channels=3,
               num_classes=3, 
               pred_patch_size=(64, 64), 
               spatial_dims=2,
               class_values=[0, 85, 170],
               task='multiclass')
    
    input_tensor = torch.randn(1, 3, 64, 64)  # Batch size 1, 3 channels, 64x64 image
    output = model.forward(input_tensor)
    
    assert output.shape == (1, 3, 64, 64), f"Unexpected output shape: {output.shape}"

    # Test for swin
    model = Unet2D(train_ds, val_ds, model='swin', 
                input_channels=3,
               num_classes=3, 
               pred_patch_size=(64, 64), 
               spatial_dims=2,
               class_values=[0, 85, 170],
               task='multiclass')
    
    input_tensor = torch.randn(1, 3, 64, 64)  # Batch size 1, 3 channels, 64x64 image
    output = model.forward(input_tensor)
    
    assert output.shape == (1, 3, 64, 64), f"Unexpected output shape: {output.shape}"


def test_training_validation_step():
    # Mock the wandb logger
    with tempfile.TemporaryDirectory() as temp_dir:
        wandb.init(dir=temp_dir, project="test_project", mode="offline")
        wandb_logger = WandbLogger(project="test_project", save_dir=temp_dir)


        train_ds = MagicMock()
        val_ds = MagicMock()
        model = Unet2D(train_ds, val_ds, model='resnet', 
                    input_channels=3,
                    num_classes=3, 
                    pred_patch_size=(64, 64), 
                    spatial_dims=2,
                    class_values=[0, 85, 170],
                    lr=3e-4,
                    batch_size=4,
                    task='multiclass',
                    num_workers=1,
                    logger=wandb_logger)
        model.has_executed = True
        model.log = MagicMock()

        batch = {"image": torch.randn(4, 3, 64, 64), "label": torch.randint(0, 3, (4, 64, 64))}
        optimizer = model.configure_optimizers()[0][0]

        optimizer.zero_grad()
        loss = model.training_step(batch, batch_idx=0)
        loss.backward()
        optimizer.step()
        assert loss.item() > 0, "Loss should be greater than 0"

        val_logs = model.validation_step(batch, batch_idx=0)
        assert "loss" in val_logs, "Validation step should log loss"

        if os.path.exists("train_tensor.pt"):
            os.remove("train_tensor.pt")
        if os.path.exists("network_graph.png"):
            os.remove("network_graph.png")


def test_prediction_function():
    train_ds = MagicMock()
    val_ds = MagicMock()
    model = Unet2D(train_ds, val_ds, model='resnet', 
                input_channels=3,
                num_classes=3, 
                pred_patch_size=(64, 64), 
                spatial_dims=2,
                class_values=[0, 85, 170],
                task='multiclass')
    
    input_image = torch.randn(1, 3, 256, 256)  
    preds = model.pred_function(input_image)
    
    assert preds.shape == (1, 3, 256, 256), f"Unexpected prediction shape: {preds.shape}"


def test_logging():
    with tempfile.TemporaryDirectory() as temp_dir:
        # wandb.login(key=os.getenv("WANDB_API_KEY"))
        wandb.init(dir=temp_dir, project="test_project", mode="offline")
        
        wandb_logger = WandbLogger(project="test_project", save_dir=temp_dir)

        model = Unet2D(train_ds=None, val_ds=None, model='resnet', 
                input_channels=3,
                num_classes=3, 
                pred_patch_size=(64, 64), 
                spatial_dims=2,
                class_values=[0, 85, 170],
                task='multiclass',
                logger=wandb_logger)
        model.log = MagicMock()

        model._compute_cnf_stats = MagicMock(return_value=(0.9, 0.85, 0.8, 0.75, 0.824))

        model.on_validation_epoch_end()

        # Check that logger.log is called with correct values
        model.log.assert_any_call('val_acc', 0.9, prog_bar=True, sync_dist=True)
        model.log.assert_any_call('val_precision', 0.85, prog_bar=False, sync_dist=True)
        model.log.assert_any_call('val_recall', 0.8, prog_bar=False, sync_dist=True)
        model.log.assert_any_call('val_iou', 0.75, prog_bar=True, sync_dist=True)
        model.log.assert_any_call('val_dice', 0.824, prog_bar=True, sync_dist=True)


if __name__ == "__main__":
    pytest.main(["-v", __file__])





import os
import json
import csv
import sys
import numpy as np
import torch
import glob
import pytorch_lightning as pl
import torchvision.utils
from torchvision.transforms import ToPILImage

from torchviz import make_dot
import neptune
from neptune.types import File

import torchmetrics
from monai.data import list_data_collate
from monai.networks.nets import UNet, SwinUNETR
from monai.networks.layers import Norm
from monai.inferers import sliding_window_inference
from monai.losses import DiceLoss

from skimage import io, color
from skimage.color import rgb2gray, rgb2lab
from skimage.morphology import disk, dilation

from PIL import Image
import torch
import pytorch_lightning as pl

from torch.utils.data import Dataset
import torchmetrics

from monai.transforms import (
    MapTransform,
    # AddChanneld, # deprecated in latest version of monai 
    # AddChannel,
    # AsChannelFirstd,
    # AsChannelLastd,
    Compose,
    CastToTyped,
    MapLabelValued,
    SqueezeDimd,
    Resized,
    CenterSpatialCropd,
    RandFlipd,
    RandAffined,
    ScaleIntensityRanged,
    EnsureTyped,
    ScaleIntensityRange,
    EnsureType,
    SpatialPadd,
    Lambda

)

# Import parent modules
try:
    from .utils import get_weights, extract_largest_component_bbox_image, MapImage
    from .utils import get_image_paths, contrast_img
except ImportError:
    from utils import get_weights, extract_largest_component_bbox_image, MapImage
    from utils import get_image_paths, contrast_img


def dynamic_scale(image: np.ndarray) -> np.ndarray:
    """Scales an input image by determining the maximum and minimum pixel values using monai's transform method ScaleIntensityRange.

    Args:
        image (np.ndarray): Input image to be scaled 

    Returns:
        np.ndarray: Scaled image.
    """
    a_min, a_max = image.min(), image.max()
    transform = ScaleIntensityRange(
        a_min=a_min,
        a_max=a_max,
        b_min=0.0,
        b_max=1.0,
        clip=True,
    )
    if image.max() > 255:
        print(image.max())
    return transform(image)


class tiff_reader(MapTransform):
    """
    Custom TIFF reader to preprocess and apply optional contrast or bounding box extraction.
    """
    def __init__(self, image_col=None, boundingbox=True, dilation=True, disk_dilation=3, keys=["image", "label"], *args, **kwargs):
        super().__init__(self, keys, *args, **kwargs)
        self.keys = keys
        self.image_col = image_col
        self.boundingbox = boundingbox
        self.dilation = dilation
        self.disk_dilation = disk_dilation

    def __call__(self, data_dict):

        data = {}
        for key in self.keys:
            if key in data_dict:
                if key == "image":
                    # if self.input_shape != img.shape[-1]
                    if self.image_col == 'cieLAB':
                        data[key] = rgb2lab(io.imread(data_dict[key]))  # cieLAB rep
                        data[key] = dynamic_scale(data[key])
                    elif self.image_col == 'grayscale':
                        data[key] = dynamic_scale(io.imread(data_dict[key]))
                        data[key] = np.expand_dims(data[key], axis=0)
                        
                    else:
                        # data[key] = np.transpose(np.array(Image.open(data_dict[key]))[..., :3], (2, 0, 1))
                        data[key] = np.array(Image.open(data_dict[key]))
                        if data[key].ndim == 4 and data[key].shape[-1] < 4:  # If shape is (h, w, d, c) assuming there are maximum 3 channels or modalities 
                            data[key] = np.transpose(data[key], (3, 0, 1, 2))  # Move channel to the first position
                            data[key] = dynamic_scale(data[key])[:3, ...]

                        elif data[key].ndim == 3 and data[key].shape[-1] <= 4:  # If shape is (h, w, c)
                            data[key] = np.transpose(data[key], (2, 0, 1))  # Move channel to the first position
                            data[key] = dynamic_scale(data[key])[:3, ...]
                        else:
                            raise ValueError(f"Unexpected image shape: {data[key].shape}, channel dimension should be last and image should be either 2D or 3D")
                elif key == "label":
                    data[key] = io.imread(data_dict[key])
                    if self.dilation:
                        data[key] = dilation(data['label'], disk(self.disk_dilation))
                    data[key] = np.expand_dims(data[key], axis=0)

        
        if self.boundingbox:
            data['image'], data['label'] = extract_largest_component_bbox_image(img=data['image'], lab=data['label'])
        return data


class ImageDataset(Dataset):
    def __init__(self, data_fnames, label_fnames, args, training=False, prediction=False):

        self.training = training
        self.prediction = prediction
        self.data_fnames = data_fnames
        self.label_fnames = label_fnames
        if len(self.data_fnames) != len(self.label_fnames):
            sys.exit(f"size of data and label images are not the same")
        self.Nsamples = len(self.data_fnames)
        self.data_dicts = [{"image": self.data_fnames[i], "label": self.label_fnames[i]} for i in range(self.Nsamples)]
        self.patch_size = args['patch_size']
        self.spatial_dims = len(self.patch_size)
        self.translate_range = args['translate_range']
        self.rotate_range = args['rotate_range']
        self.scale_range = args['scale_range']
        self.shear_range = args['shear_range']
        self.class_values = args['class_values']
        self.image_col = args["image_col"]
        self.input_channels = args['input_channels']
        self.target_size = args['patch_size']
        self.boundingbox = args['boundingbox']
        self.dilation = args['dilation']
        self.disk_dilation = args['disk_dilation']

        self.transform = self.get_data_transforms(self.training, self.boundingbox, self.dilation, self.disk_dilation)

    def __len__(self):
        return self.Nsamples

    def __getitem__(self, idx):
        if self.prediction: 
            return (self.transform(self.data_dicts[idx]), os.path.split(self.data_fnames[idx])[-1])
        else:
            return self.transform(self.data_dicts[idx])

    # These transforms define the data preprocessing and augmentations done to the raw images BEFORE input to neural network
    def get_data_transforms(self, training, boundingbox, dilation, disk_dilation):
        if not training:  # validation or test_masks set -- no image augmentation
            transform = Compose(
                [
                    tiff_reader(keys=["image", "label"], image_col=self.image_col, boundingbox=boundingbox, dilation=dilation, disk_dilation=disk_dilation),
                    Resized(keys=["image", "label"], spatial_size=self.target_size, mode=['area', 'nearest']),
                    MapLabelValued(["label"],
                                    self.class_values,
                                    list(range(len(self.class_values)))),
                    SqueezeDimd(keys=["label"], dim=0),
                    CastToTyped(keys=["label"], dtype=torch.long),
                    EnsureTyped(keys=["image", "label"])
                ]
            )
            
        else:  # training set -- do image augmentation
            transform = Compose(
                [
                    tiff_reader(keys=["image", "label"], image_col=self.image_col, boundingbox=boundingbox, dilation=dilation, disk_dilation=disk_dilation),
                    Resized(keys=["image", "label"], spatial_size=self.target_size, mode=['area', 'nearest']),
                    RandFlipd(
                        keys=['image', 'label'],
                        prob=0.5,
                        spatial_axis=(0, 1)
                    ),
                    RandAffined(
                        keys=['image', 'label'],
                        mode=['bilinear', 'nearest'],
                        padding_mode='zeros',  ### implicitly assumes raw unlabeled_index = 0!
                        prob=1.0,
                        spatial_size=self.patch_size,
                        rotate_range=self.rotate_range * np.ones(1),
                        translate_range=self.translate_range * np.asarray(self.patch_size),
                        shear_range=self.shear_range * np.ones(2),
                        scale_range=self.scale_range * np.ones(2)
                    ),
                    MapLabelValued(["label"],
                                    self.class_values,
                                    list(range(len(self.class_values)))),
                    SqueezeDimd(["label"], dim=0),
                    CastToTyped(keys=["label"], dtype=torch.long),
                    EnsureTyped(keys=["image", "label"])
                ]
            )
           
        return transform


class PredDataset2D(Dataset):
    def __init__(self, pred_data_dir, args):
        self.pred_data_dir = pred_data_dir
        self.data_file = get_image_paths(pred_data_dir)
        # self.data_file = pred_data_lst
        self.input_col = args['input_channels']
        self.image_col = args['image_col']
        self.boundingbox = args['boundingbox']

    def __len__(self):
        return len(self.data_file)

    def __getitem__(self, idx):
        
        img_name = self.data_file[idx]
        img_path = os.path.join(self.pred_data_dir, img_name)
        # img_path = self.data_file[idx]
        img = io.imread(img_path)
        if img.ndim == 4 and img.shape[-1] < 4:  # If shape is (h, w, d, c) assuming there are maximum 3 channels or modalities 
            img = np.transpose(img[...,:3], (3, 0, 1, 2))  # Move channel to the first position
            img = dynamic_scale(img)
        elif img.ndim == 3 and img.shape[-1] <= 4:  # If shape is (h, w, c)
            img = np.transpose(img[...,:3], (2, 0, 1))  # Move channel to the first position
            img = dynamic_scale(img)
        elif self.input_col == 1: # grayscale image 
            img = dynamic_scale(img)
            img = np.expand_dims(img, axis=0)

        else:
            raise ValueError(f"Unexpected image shape: {img.shape}, channel dimension should be last and image should be either 2D or 3D")
        
        transform = Compose(
            [
                EnsureType()
            ]
        )
        img = transform(img)
        if self.boundingbox:
            img = extract_largest_component_bbox_image(img)
        return (img, img_path)


class Unet2D(pl.LightningModule):
    def __init__(self, train_ds, val_ds, **kwargs):
        """    
        PyTorch Lightning Module for training and evaluating 2D and 3D UNet-based models.
        This module supports both the traditional UNet and SwinUNETR architectures 
        for image segmentation tasks.

        Args:
            train_ds (Dataset): Training dataset
            val_ds (Dataset): Validation dataset

        Example: 

        >>> from monai.networks.nets import UNet, SwinUNETR

        >>> # 5 layers each down/up sampling their inputs by a factor 2 with residual blocks
        >>> model = UNet(
        >>>     spatial_dims=2,
        >>>     in_channels=3,
        >>>     out_channels=3,
        >>>     channels=(32, 64, 128, 256, 512),
        >>>     strides=(2, 2, 2, 2),
        >>>     kernel_size=3,
        >>>     up_kernel_size=3,
        >>>     num_res_units=2,
        >>>     dropout=0.4,
        >>>     norm=Norm.BATCH,
        >>>     )

        >>> # SwinUNETR model with 3D inputs and batch normalization
        >>> model = SwinUNETR(
        >>>     spatial_dims=3,
        >>>     img_size=(256, 256, 256),
        >>>     in_channels=3,
        >>>     out_channels=3,
        >>>     feature_size=48,
        >>>     num_heads=[3, 6, 12, 24, 12] ,
        >>>     depths=[2, 4, 8, 16, 24] ,
        >>>     drop_rate=0.4,
        >>>     attn_drop_rate=0.2,
        >>>     norm_name='batch'
        >>>     )
        """
        super(Unet2D, self).__init__()

        self.save_hyperparameters()
        self.hparams.output_channels = self.hparams.num_classes
        self.hparams.pred_patch_size = self.hparams.pred_patch_size
        self.hparams.model = self.hparams.model
        self.hparams.spatial_dims = self.hparams.spatial_dims
        self.hparams.class_values = self.hparams.class_values
        self.train_ds = train_ds
        self.val_ds = val_ds
        self.hparams.task = 'multiclass'

        self.cnfmat = torchmetrics.ConfusionMatrix(num_classes=self.hparams.num_classes,
                                                   task=self.hparams.task,
                                                   normalize=None)
        if self.hparams.model == 'resnet':
            self.model = UNet(
                spatial_dims=self.hparams.spatial_dims,
                in_channels=self.hparams.input_channels,
                out_channels=self.hparams.output_channels,
                channels=(32, 64, 128, 256, 512),
                strides=(2, 2, 2, 2),
                kernel_size=3,
                up_kernel_size=3,
                num_res_units=2,
                dropout=0.4,
                norm=Norm.BATCH,
            )

        
        elif self.hparams.model == "swin":
            img_size = self.hparams.pred_patch_size
            feature_size = 48        
            use_checkpoint = True   
            dropout_rate = 0.4
            attention_dropout_rate = 0.2
            depths = [2, 4, 8, 16, 24]  
            num_heads = [3, 6, 12, 24, 12] 

            # Instantiate the model
            self.model = SwinUNETR(
                spatial_dims=self.hparams.spatial_dims,
                img_size=img_size,
                in_channels=self.hparams.input_channels,
                out_channels=self.hparams.output_channels,
                feature_size=feature_size,
                num_heads=num_heads,
                depths=depths,
                use_checkpoint=use_checkpoint,
                drop_rate=dropout_rate,
                attn_drop_rate=attention_dropout_rate,
                norm_name='batch'
                # norm_name = ('group', {"num_groups": 3, "affine": True})
            )
        else:
            raise AttributeError ("Only 2 possibles models can be applied, either resnet or swin models")
        self.has_executed = False

    @staticmethod
    def add_args(parser):
        parser.add_argument("--num_classes", type=int, default=3, help="number of segmentation classes to predict")
        parser.add_argument("--input_channels", type=int, default=3,
                            help="number of input channels (1 for grayscale, 3 for RGB)")
        parser.add_argument("--background_index", type=int, default=0, help="background index")
        parser.add_argument("--pred_patch_size", type=int, default=(64, 64),
                            help="spatial size of rolling window prediction patch")
        parser.add_argument("--batch_size", type=int, default=1, help="dataloader batch size")
        parser.add_argument("--lr", type=int, default=3e-4, help="Adam learning rate")
        parser.add_argument("--num_workers", type=int, default=4, help="number of dataloader workers (cpus)")
        return parser

    def forward(self, x):
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        return self.model(x.to(device))

    def train_dataloader(self):
        """
        Creates and returns the DataLoader for the training dataset.

        Returns:
            DataLoader: A PyTorch DataLoader for the training dataset.
        """
        train_loader = torch.utils.data.DataLoader(
            self.train_ds, batch_size=self.hparams.batch_size, shuffle=True,
            collate_fn=list_data_collate, num_workers=self.hparams.num_workers,
            persistent_workers=True, pin_memory=torch.cuda.is_available())
        return train_loader

    def val_dataloader(self):
        """
        Creates and returns the DataLoader for the validation dataset.

        Returns:
            DataLoader: A PyTorch DataLoader for the validation dataset.
        """
        val_loader = torch.utils.data.DataLoader(
            self.val_ds, batch_size=self.hparams.batch_size, shuffle=False,
            collate_fn=list_data_collate, num_workers=self.hparams.num_workers,
            persistent_workers=True, pin_memory=torch.cuda.is_available())
        return val_loader

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.model.parameters(), self.hparams.lr, weight_decay=0.001)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer,
                                                               mode='min',
                                                               factor=0.2,
                                                               patience=10,
                                                               min_lr=1e-6,
                                                               verbose=True)
        lr_scheduler = {"scheduler": scheduler,
                        "monitor": "val_loss",
                        "interval": "epoch"}

        return [optimizer], [lr_scheduler]

    def loss_function(self, logits, labels, class_weights=None):
        cross_entropy_loss = torch.nn.CrossEntropyLoss(weight=class_weights, reduction='mean')
        loss = cross_entropy_loss(logits, labels)

        # dice_loss = DiceLoss(softmax=True, include_background=True, weight=class_weights, to_onehot_y=True, reduction='mean')
        # loss = dice_loss(logits, labels)
        return loss

    def training_step(self, batch, batch_idx):
        images, labels = batch["image"], batch["label"]
        logits = self.forward(images)  # forward pass on a batch
        classes = list(range(self.hparams.num_classes))
        # torch.save(images, 'train_tensor.pt')

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        class_weights = get_weights(labels, classes, device, include_background=True)

        loss = self.loss_function(logits, labels, class_weights)
        self.loss = loss.item()
        self.log("train_loss", loss.item(), on_step=False, on_epoch=True, prog_bar=True)

        # if not self.has_executed:
        #     dummy_input = torch.rand((1, self.hparams.input_channels,) + self.hparams.pred_patch_size) 
        #     make_dot(self(dummy_input), params=dict(self.named_parameters())).render('network_graph', format='png')
        #     self.logger.log_image(key="model_visualization", images=["network_graph.png"]) # Wandb Logger
        #     self.has_executed = True
            
        # if os.path.exists("network_graph.png"):
        #     os.remove("network_graph.png")
        # if os.path.exists("train_tensor.pt"):
        #     os.remove("train_tensor.pt")

        return loss

    def validation_step(self, batch, batch_idx):
        images, labels = batch["image"], batch["label"]
        logits = self.forward(images)
        classes = list(range(self.hparams.num_classes))
        
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        
        class_weights = get_weights(labels, classes, device, include_background=True)
        loss = self.loss_function(logits, labels, class_weights)
        
        labeled_ind = (labels != -1)
        labeled_logits = torch.cat([logits[i, :, labeled_ind[i, :, :]] for i in range(len(logits))], dim=-1).transpose(
            0, 1)

        batch_logs = {"loss": loss,
                      "logits": logits,
                      "labels": labels}
        if not self.has_executed:
            to_pil = ToPILImage()
            if batch_idx % 100 == 0:
                x, y = images[:20], logits[:20]
                gridx = torchvision.utils.make_grid(x, nrow=5)

                # Convert the grid to a PIL image
                self.logger.log_image(key="training_img", images=[to_pil(gridx)]) # Wandb Logger
                preds = torch.argmax(y, dim=1).byte().squeeze(1)
                preds = (preds * 255).byte()


                y = MapImage(preds, self.hparams.class_values, reverse=False)
                
                gridy = torchvision.utils.make_grid(y.view(y.shape[0], 1, y.shape[1], y.shape[2]), nrow=5)
                self.logger.log_image(key="prediction_imgs", images=[to_pil(gridy)]) # Wandb Logger

        if torch.cuda.is_available():
            self.cnfmat(labeled_logits, labels[labeled_ind])
        else:
            self.cnfmat(labeled_logits, torch.tensor(labels)[torch.tensor(labeled_ind)])
        self.log("val_loss", loss.item(), on_step=False, on_epoch=True, prog_bar=True, sync_dist=True)
        return batch_logs

    def on_validation_epoch_end(self):
        acc, prec, recall, iou, dice = self._compute_cnf_stats()

        self.log('val_acc', acc, prog_bar=True, sync_dist=True)
        self.log('val_recall', recall, prog_bar=False, sync_dist=True)
        self.log('val_precision', prec, prog_bar=False, sync_dist=True)
        self.log('val_iou', iou, prog_bar=True, sync_dist=True)
        self.log('val_dice', dice, prog_bar=True, sync_dist=True)

        val_logs = {'log':
                        {'val_acc': acc,
                         'val_recall': recall,
                         'val_precision': prec,
                         'val_iou': iou,
                         'val_dice': dice}
                    }

        return val_logs

    def test_step(self, batch, batch_idx):
        images, labels = batch["image"], batch["label"]
        logits = self.forward(images)
        labeled_ind = (labels != -1)
        labeled_logits = torch.cat([logits[i, :, labeled_ind[i, :, :]] for i in range(len(logits))], dim=-1).transpose(
            0, 1)

        if torch.cuda.is_available():
            self.cnfmat(labeled_logits, labels[labeled_ind])
        else:
            self.cnfmat(labeled_logits, torch.tensor(labels.numpy())[torch.tensor(labeled_ind.numpy())])

    def on_test_epoch_end(self):
        acc, prec, recall, iou, dice = self._compute_cnf_stats()

        print(f"test performance: acc={acc:.02f}, precision={prec:.02f}, recall={recall:.02f}, iou={iou:.02f}, dice={dice:.02f}")
        self.log('test_acc', acc, sync_dist=True)
        self.log('test_recall', recall, sync_dist=True)
        self.log('test_precision', prec, sync_dist=True)
        self.log('test_iou', iou, sync_dist=True)
        self.log('test_dice', dice, sync_dist=True)

        with open(os.path.join(self.hparams.log_dir, 'test_stats.csv'), 'w') as f:
            writer = csv.writer(f)
            # writer.writerow(["acc, prec, recall, iou, dice"])
            # writer.writerow([f"{acc:.02f}", f"{prec:.02f}", f"{recall:.02f}", f"{iou:.02f}", f"{dice:.02f}"])
            writer.writerow(["acc", "prec", "recall", "iou", "dice"])
            writer.writerow([acc, prec, recall, iou, dice])


    def pred_function(self, image):
        return sliding_window_inference(image, self.hparams.pred_patch_size, 1, self.forward)

    def predict_step(self, batch, batch_idx):
        images, fnames = batch
        # cropped_images = extract_largest_component_bbox_image(images)
        # tensor_cropped_images = torch.tensor(cropped_images).to("cuda")
        logits = self.pred_function(images.float())
        preds = torch.argmax(logits, dim=1).byte().squeeze(dim=1)
        images = (images * 255).byte()  # convert from float (0-to-1) to uint8
        return (preds, images, fnames)


# The following code assign compute true negative s.t. true positives of the background as the true negatives for all other classes
# This works only in binary settings where the background class is the only negative class 
# For general multiclass tasks, we need a more robust calculation that sums over all non-class rows and columns to avoid misrepresenting the true negatives.
    def _compute_cnf_stats(self):
        cnfmat = self.cnfmat.compute()
        true = torch.diag(cnfmat).as_tensor() 
        tn = true[self.hparams.background_index]
        tp = torch.cat([true[:self.hparams.background_index], true[self.hparams.background_index + 1:]])

        fn = (cnfmat.sum(1) - true)[torch.arange(cnfmat.size(0)) != self.hparams.background_index]
        fp = (cnfmat.sum(0) - true)[torch.arange(cnfmat.size(1)) != self.hparams.background_index]

        acc = torch.sum(true) / torch.sum(cnfmat)
        precision = torch.sum(tp) / torch.sum(tp + fp)
        recall = torch.sum(tp) / torch.sum(tp + fn)
        iou = torch.sum(tp) / (torch.sum(cnfmat) - tn)
        # dice = 2*torch.sum(2*tp) / (torch.sum(cnfmat) + tp - tn)
        dice = 2 * torch.sum(tp) / (2 * torch.sum(tp) + torch.sum(fp) + torch.sum(fn))

        iou_per_class = tp / (tp + fp + fn)

        self.cnfmat.reset()

        return acc.mean().item(), precision.mean().item(), recall.mean().item(), iou.mean().item(), dice.mean().item()


    # Compute metrics without excluding background
    # def _compute_cnf_stats(self):
    #     cnfmat = self.cnfmat.compute()  # Confusion matrix: shape (num_classes, num_classes)
    #     true = torch.diag(cnfmat)  # True positives for each class

    #     # Total true positives, false negatives, and false positives for each class
    #     tp = true
    #     fn = cnfmat.sum(dim=1) - true  # Row sum minus true positives
    #     fp = cnfmat.sum(dim=0) - true  # Column sum minus true positives
    #     tn = cnfmat.sum() - (tp + fn + fp)  # Total - (tp, fp, fn)

    #     # Metrics
    #     acc = torch.sum(tp) / torch.sum(cnfmat)  # Overall accuracy
    #     precision = torch.mean(tp / (tp + fp + 1e-8))  # Macro-averaged precision
    #     recall = torch.mean(tp / (tp + fn + 1e-8))  # Macro-averaged recall
    #     iou = torch.mean(tp / (tp + fp + fn + 1e-8))  # Macro-averaged IoU
    #     iou_per_class = tp / (tp + fp + fn + 1e-8)  # Per-class IoU

    #     self.cnfmat.reset()  # Reset confusion matrix for next evaluation cycle

    #     return acc.item(), precision.item(), recall.item(), iou.item()



import os
import torch
import numpy as np
from unittest.mock import MagicMock, patch
from skimage import io
from PIL import Image
import pytest
from torchvision.transforms import ToPILImage
import sys


if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")


from rhizonet.predict import transform_image, predict_step, get_prediction, predict_model, pred_function
from rhizonet.unet2D import Unet2D


def test_transform_image():
    """Test the image transformation logic."""
    test_img_path = "test_image.png"
    test_img = np.random.randint(0, 255, (256, 256, 3), dtype=np.uint8)
    Image.fromarray(test_img).save(test_img_path)

    img, img_path = transform_image(test_img_path)

    assert img.shape == (3, 256, 256), "Image transformation failed."
    assert img_path == test_img_path, "Image path mismatch."

    os.remove(test_img_path)


def test_get_prediction():
    """Test the binary mask generation and saving."""
    test_file = "test_image.png"
    save_path = "test_output"
    os.makedirs(save_path, exist_ok=True)

    topil = ToPILImage()
    test_img = torch.rand(3, 256, 256, dtype=torch.float32)
    test_img = topil(test_img)
    test_img.save(test_file)

    mock_model = MagicMock()
    mock_model.return_value = torch.randn(1, 3, 128, 128)  # Mock logits output

    get_prediction(test_file, mock_model, (128, 128), save_path, labels=[1, 0], binary_preds=True, frg_class=255)

    saved_file = os.path.join(save_path, os.path.basename(test_file).split('.')[0] + ".png")
    assert os.path.exists(saved_file), "Prediction output not saved."

    os.remove(test_file)
    os.remove(saved_file)
    os.rmdir(save_path)

def test_get_prediction_grayscale():
    """Test the binary mask generation and saving."""
    test_file = "test_image.png"
    save_path = "test_output"
    os.makedirs(save_path, exist_ok=True)

    topil = ToPILImage()
    test_img = torch.rand(256, 256, dtype=torch.float32)
    test_img = topil(test_img)
    test_img.save(test_file)

    mock_model = MagicMock()
    mock_model.return_value = torch.randn(1, 1, 128, 128)  # Mock logits output

    get_prediction(test_file, mock_model, (128, 128), save_path, labels=[1, 0], binary_preds=True, frg_class=255)

    saved_file = os.path.join(save_path, os.path.basename(test_file).split('.')[0] + ".png")
    assert os.path.exists(saved_file), "Prediction output not saved."

    os.remove(test_file)
    os.remove(saved_file)
    os.rmdir(save_path)



if __name__ == "__main__":
    pytest.main(["-v", __file__])



"""
Script for training and evaluation a residual U-Net model for image segmentation.
The training process of RhizoNet includes logging, checkpointing and metrics evaluation.

Dependencies:
- PyTorch Lightning
- MONAI
- Scikit-image
- wandb

Usage:
    pip install rhizonet
    train_rhizonet --config_file ./setup_files/setup-train.json --gpus 2 --strategy "ddp" --accelerator "gpu"
"""

import os
import json
import csv
import numpy as np
from argparse import ArgumentParser
import torch
import glob
from tqdm import tqdm
import pytorch_lightning as pl
from skimage import io, color
from argparse import Namespace
from pathlib  import Path

from monai.data import list_data_collate
from lightning.pytorch.loggers import WandbLogger
import wandb


# Import parent modules
try:
    from .utils import MapImage, createBinaryAnnotation, get_image_paths
    from .metrics import evaluate
    from .unet2D import Unet2D, ImageDataset, PredDataset2D
    from .simpleLogger import mySimpleLogger
except ImportError:
    from utils import MapImage, createBinaryAnnotation, get_image_paths
    from metrics import evaluate
    from unet2D import Unet2D, ImageDataset, PredDataset2D
    from simpleLogger import mySimpleLogger



def _parse_training_variables(argparse_args):
    """ 
    Parse and merge training variables from a JSON configuration file and command-line arguments.

    Args:
        argparse_args (Namespace): Command-line arguments parsed by argparse.

    Returns:
        tuple: Updated arguments, dataset parameters, and model parameters. 
    """
    args = vars(argparse_args)

    # overwrite argparse defaults with config file
    with open(args["config_file"]) as file_json:
        config_dict = json.load(file_json)
        args.update(config_dict)

    dataset_args, model_args = args['dataset_params'], args['model_params']
    dataset_args['patch_size'] = tuple(dataset_args['patch_size'])  # tuple expected, not list
    model_args['pred_patch_size'] = tuple(model_args['pred_patch_size'])  # tuple expected, not list

    return args, dataset_args, model_args


def train_model(args):
    """
    Train and evaluate RhizoNet on a specified dataset.

    Args:
        args (Namespace): Command-line containing: 
            - config_file (str): Path to the JSON Configuration of the model.
            - gpus (int): Number of gpu nodes to use training.
            - strategy (str): Strategy to use for training (e.g., 'ddp', 'dp')
            - accelerator (str): cpu or gpu training
    
    Returns: None

    Notes:
        - The evaluation results are saved to the file specified by 'save_path' in the configuration file 
        - Training and validation metrics are also available in the WandDB project 'rhizonet'
        - Predictions associated to the full size images specified in 'pred_data_dir' are generated after training and saved in the 'save_path' directory.
        - Metrics (accuracy, precision, recall and IOU) are evaluated on full size images specified in 'pred_data_dir' and results are saved in a metrics.json files 

    Example::
        Run this script using the following command-line if 2 GPU nodes available: 
            python train.py --gpus 2 --strategy "ddp" --config_file "./setup_files/setup-train.json"
        
        Run this script using the following command-line if 1 GPU node available: 
            python train.py --gpus 1 --strategy "dp" --config_file "./setup_files/setup-train.json"
    """


    # Load image and label filepaths 
    args, dataset_params, model_params = _parse_training_variables(args)
    image_dir, label_dir, log_dir = model_params['image_dir'], model_params['label_dir'], model_params['log_dir']
    os.makedirs(log_dir, exist_ok=True)
    
    # Save json parameters to log directory
    with open(os.path.join(log_dir, 'training_parameters.json'), 'w') as f:
        json.dump(args, f)

    images, labels = [], []

    images = get_image_paths(image_dir)
    labels = get_image_paths(label_dir)

#    Get matching images and labels 
    lbl_extension = Path(labels[0]).suffix
    i = 0 # count number of images removed that did not appear in list of labels
    print("Checking that all filepaths in images have matching label")
    for img_path in tqdm(images):
        img_name = os.path.splitext(img_path)[0].split("/")[-1]
        lbl_name = dataset_params["label_prefix"] + img_name + lbl_extension
        if not os.path.isfile(os.path.join(label_dir, lbl_name)):
            images.remove(img_path)
            i += 1
    print("Removed {} images from the list of images after match checking".format(i))

    # Split data into training, validation and test sets
    train_len, val_len, test_len = np.cumsum(np.round(len(images) * np.array(dataset_params['data_split'])).astype(int))
    idx = np.random.permutation(np.arange(len(images)))

    train_images = [images[i] for i in idx[:train_len]]
    train_labels = [labels[i] for i in idx[:train_len]]
    val_images = [images[i] for i in idx[train_len:val_len]]
    val_labels = [labels[i] for i in idx[train_len:val_len]]
    test_images = [images[i] for i in idx[val_len:]]
    test_labels = [labels[i] for i in idx[val_len:]]
    
    # Create datasets
    train_dataset = ImageDataset(train_images, train_labels, dataset_params, training=True)
    val_dataset = ImageDataset(val_images, val_labels, dataset_params, )
    test_dataset = ImageDataset(test_images, test_labels, dataset_params, )

    # Initialize Lightning module
    unet = Unet2D(train_dataset, val_dataset, **model_params)

    # Set up logging and callbacks
    wandb.login(key=os.getenv("WANDB_API_KEY"))
    wandb_logger = WandbLogger(log_model="all",
                               project="rhizonet")

    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        dirpath=log_dir,
        filename="checkpoint-{epoch:02d}-{val_loss:.2f}",
        save_top_k=1,
        every_n_epochs=1,
        save_weights_only=True,
        verbose=True,
        monitor="val_acc",
        mode='max')
    stopping_callback = pl.callbacks.EarlyStopping(monitor='val_loss',
                                                   min_delta=1e-3,
                                                   patience=10,
                                                   verbose=True,
                                                   mode='min')
    lr_monitor = pl.callbacks.LearningRateMonitor(logging_interval='epoch', log_momentum=False)

    # Initialize PyTorch Lightning trainer
    trainer = pl.Trainer(
        # precision="16-mixed", 
        # accumulate_grad_batches=2, # for batch size 2, add these 2 lines if memory issues
        default_root_dir=log_dir,
        callbacks=[checkpoint_callback, lr_monitor, stopping_callback],
        log_every_n_steps=1,
        enable_checkpointing=True,
        logger=wandb_logger,
        accelerator=args['accelerator'],
        devices=args['gpus'],
        strategy=args['strategy'],
        num_sanity_val_steps=0,
        max_epochs=model_params['nb_epochs']
    )

    # Train the model
    trainer.fit(unet)

    # Test the model
    test_loader = torch.utils.data.DataLoader(
        test_dataset, batch_size=model_params['batch_size'], shuffle=False,
        collate_fn=list_data_collate, num_workers=model_params["num_workers"],
        persistent_workers=True, pin_memory=torch.cuda.is_available())
    trainer.test(unet, test_loader, ckpt_path='best', verbose=True)

    # Generate predictions
    pred_img_path = os.path.join(model_params['pred_data_dir'], "images")
    pred_lab_path = os.path.join(model_params['pred_data_dir'], "labels")
    predict_dataset = PredDataset2D(pred_img_path, dataset_params)
    predict_loader = torch.utils.data.DataLoader(
        predict_dataset, batch_size=1, shuffle=False,
        collate_fn=list_data_collate, num_workers=model_params["num_workers"],
        persistent_workers=True, pin_memory=torch.cuda.is_available())
    
    predictions = trainer.predict(unet, predict_loader)
    
    # Save predictions
    pred_path = os.path.join(log_dir, 'predictions')
    os.makedirs(pred_path, exist_ok=True)

    labels = dataset_params['class_values']
    for (pred, _, fname) in predictions:
        pred = MapImage(pred, labels, reverse=True)
        pred = pred.numpy().squeeze().astype(np.uint8)
        fname = os.path.basename(fname[0]).split('.')[0] + ".png"
        # pred_img, mask = elliptical_crop(pred, 1000, 1500, width=1400, height=2240)
        if dataset_params['binary_preds']:
            binary_mask = createBinaryAnnotation(pred, dataset_params['frg_class']).astype(np.uint8)
            io.imsave(os.path.join(pred_path, fname), binary_mask, check_contrast=False)
        else:
            io.imsave(os.path.join(pred_path, fname), pred, check_contrast=False)


    # Evaluate metrics on full size test images 
    if dataset_params['binary_preds']:
        evaluate(pred_path, pred_lab_path, log_dir, task='binary', num_classes=2, frg_class = dataset_params['frg_class'])
    else:
        evaluate(pred_path, pred_lab_path, log_dir, task='multiclass', num_classes=len(labels), frg_class = dataset_params['frg_class'])
        

def main():
    parser = ArgumentParser(conflict_handler='resolve')
    parser.add_argument("--config_file", type=str,
                        default="../data/setup_files/setup-train.json",
                        help="json file training data parameters")
    parser.add_argument("--gpus", type=int, default=1, help="how many gpus to use")
    parser.add_argument("--strategy", type=str, default='ddp', help="pytorch strategy")
    parser.add_argument("--accelerator", type=str, default='gpu', help="cpu or gpu accelerator")

    args = parser.parse_args()

    train_model(args)

if __name__ == "__main__":
    main()




import numpy as np
import torch
import torchvision.transforms.functional as TF
from PIL import Image, ImageDraw
import os
from skimage.color import rgb2hsv
from skimage import exposure
from skimage import io, filters, measure
from scipy import ndimage as ndi
from typing import Union, List, Tuple, Sequence, Dict
from monai.transforms import MapLabelValued


def extract_largest_component_bbox_image(img: Union[np.ndarray, torch.Tensor], 
                                         lab: Union[np.ndarray, torch.Tensor] = None,) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Extract the largest connected component (LCC) of the given image and get the bounding box associated with this LCC. Depending on the input parameters, it can either:

    - Crop the image to the bounding box.
    - Apply a binary mask to the image, maintaining the same shape as the input image, with everything but the LCC set as a black background.

    This function can be applied to both the image and the annotation (label) if provided.

    Args:
        img (Union[numpy.ndarray, torch.Tensor]): 
            The input image. 
            - Can be a NumPy array or PyTorch tensor with shape (C, H, W) or (B, C, H, W) if a batch dimension is included.
            - Must not be None; if None, the function raises a ValueError.
        lab (Union[numpy.ndarray, torch.Tensor], optional): 
            The label or annotated image. Defaults to None.


    Returns:
        Union[numpy.ndarray, Tuple[numpy.ndarray, numpy.ndarray]]: 
            - If `lab` is None: Returns the processed image as a NumPy array or PyTorch tensor.
            - If `lab` is provided: Returns a tuple containing:
                - The processed image (NumPy array).
                - The processed label image (NumPy array).
    """

    if img is None:
        raise ValueError("Image cannot be None.")
    elif isinstance(img, np.ndarray):
        print("Processing a NumPy array.")
    elif isinstance(img, torch.Tensor):
        print("Processing a PyTorch tensor")
        img = np.array(img)
    else:
        raise TypeError("Input must be a numpy.ndarray, torch.Tensor, or None.")
    
    # Remove dimension if there is a batch dim and format is (B, C, H, W) or (B, H, W) for grayscale images
    if len(img.squeeze().shape) >= 3 and img.squeeze().shape[0]>=3: #RGB or RGBA images after removing the batch dimension
        image = img.squeeze()
        image = image[2, ...] # choose a channel 
    else:
        image = img.squeeze()
        
    # Get the largest connected component
    image = ndi.gaussian_filter(image, sigma=2)

    # Threshold the image
    threshold = filters.threshold_isodata(image)
    binary_image = image < threshold

    # Label connected components
    label_image = measure.label(binary_image)
    # Measure properties of the connected components
    props = measure.regionprops(label_image)

    # Find the largest connected component by area
    if props:
        largest_component = max(props, key=lambda x: x.area)
        largest_component_mask = label_image == largest_component.label
    else:
        largest_component_mask = np.zeros_like(binary_image, dtype=bool)

    # Fill all holes in the largest connected component
    filled_largest_component_mask = ndi.binary_fill_holes(largest_component_mask)

    # Get the bounding box of the largest connected component
    min_row, min_col, max_row, max_col = largest_component.bbox

    # Crop the ORIGINAL image to the bounding box dimensions
    cropped_image = img[..., min_row:max_row, min_col:max_col]

    # Create a new image with the cropped content but keeping input image shape
    new_image = np.zeros_like(img)

    # Applying the mask without cropping out the ROI 
    new_image[..., min_row:max_row, min_col:max_col] = cropped_image * filled_largest_component_mask[min_row:max_row, min_col:max_col]

    # Processing the label image
    if lab is not None:
        cropped_label = lab[..., min_row:max_row, min_col:max_col]
        new_label = np.zeros_like(lab)
        new_label[..., min_row:max_row, min_col:max_col] = cropped_label * filled_largest_component_mask[min_row:max_row, min_col:max_col]
        return new_image, new_label
    
    return new_image
        

def get_weights(
        labels: torch.Tensor, 
        classes: List[int], 
        device: str, 
        include_background=False,
        ) -> List[float]:
    """
    Computes the weights of each class in the batch of labeled images 

    Args:
        labels (torch.Tensor): Batch of labeled imagse with each pixel of an image equal to a class value. 
        classes (List[int]): List of the classes
        device (str): training device should be 'cuda' if training on GPU
        include_background (bool, optional): Boolean to include or not the background valued as 0 when calculating the weight of the classes in the images. Defaults to False.

    Returns:
        List[float]: List of the class weights (floats).
    """

    labels = labels.to(device)
    if not include_background:
        classes.remove(0)
    flat_labels = labels.view(-1)

    # FIX IF NOT ALL CLASSES ARE IN THE LABEL INPUT 
    n = len(classes)
    class_counts = torch.bincount(flat_labels, minlength=n)
    class_weights = torch.zeros(n, dtype=torch.float, device=device)
    nonzero_mask = class_counts > 0
    class_weights[nonzero_mask] = 1 / class_counts[nonzero_mask]
    class_weights /= class_weights.sum()
    # print("class weights {}".format(class_weights))

    return class_weights


def MapImage(
        image: Union[np.ndarray, torch.Tensor], 
        original_values: List[int],
        reverse: bool
        ) -> Union[np.ndarray, torch.Tensor]:
    """
    Maps the current values of a given input image to the values given by the tuple (current values, new values).

    Args:
        image (Union[np.ndarray, torch.Tensor]): The input image to transform
        original_values (List[int]): List of original values to be mapped
        reverse (bool): True if mapping the other way around back to original values

    Raises:
        TypeError: If the input image is neither a numpy array or a torch tensor

    Returns:
        Union[np.ndarray, torch.Tensor]: the transformed input after mapping.
    
    Example::
        transformed_image = MapImage(image, [0, 85, 170])
        The values will be mapped to [0, 1, 2] with the first value of the original values specified being the background index. 
    """
    if isinstance(image, np.ndarray) :      
        data = image.copy()
    elif isinstance(image, torch.Tensor):
        data = image.detach()
    else:
        raise TypeError("Input must be a numpy.ndarray, torch.Tensor")
    
    target_values = list(range(len(original_values)))
    if not reverse:
        # Create the transform
        map_label_transform = MapLabelValued(["label"], original_values, target_values)
    else:
        # Create the transform
        map_label_transform = MapLabelValued(["label"], target_values, original_values)
        
    # Apply the transform
    mapped_label_image = map_label_transform({"label": data})["label"]

    return mapped_label_image


def elliptical_crop(img: np.ndarray, 
                    center_x: int, center_y: int, 
                    width: int, 
                    height: int
                    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Crops out an elliptical shape out of the input image and sets the rest as background 

    Args:
        img (np.ndarray): Input image
        center_x (int): Center x coordinate of the ellipse 
        center_y (int): Center y coordinate of the ellipse 
        width (int): Width of the wanted ellipse
        height (int): Height of the wanted ellipse

    Returns:
        Tuple[np.ndarray, np.ndarray]: cropped output image
    """

    image = Image.fromarray(img)
    image_width, image_height = image.size

    # Create an elliptical mask using PIL
    mask = Image.new('1', (image_width, image_height), 0)
    draw = ImageDraw.Draw(mask)
    draw.ellipse((center_x - width / 2, center_y - height / 2, center_x + width / 2, center_y + height / 2), fill=1)

    # Convert the mask to a PyTorch tensor
    mask_tensor = TF.to_tensor(mask)

    # Apply the mask to the input image using element-wise multiplication
    cropped_image = TF.to_pil_image(torch.mul(TF.to_tensor(image), mask_tensor))

    return image, np.array(cropped_image)



def get_image_paths(dir: str) -> List[str]:
    """
    Goes through a folder directory and lists all filepaths

    Args:
        dir (str): folder directory to extract from all the filepaths 

    Returns:
        List[str]: list of all filepaths
    """

    image_files = []
    for root, directories, files in os.walk(dir):
        for filename in files:
            if not filename.startswith("."):
                image_files.append(os.path.join(root, filename))  
    return sorted(image_files)


def contrast_img(img: np.ndarray) -> np.ndarray:
    """
    Applies the Adaptive Equalization or histogram equalization contrast method. This method
    enhances the contrast of an image by adjusting the intensity values of pixels based on the 
    distribution of pixel intensities in the image's histogram. 

    Args:
        img (np.ndarray): input image

    Returns:
        np.ndarray: contrasted image
    """
    # HSV image
    hsv_img = rgb2hsv(img)  # 3 channels
    # select 1channel
    img = hsv_img[:, :, 0]
    # Contrast stretching
    p2, p98 = np.percentile(img, (2, 98))
    img = exposure.rescale_intensity(img, in_range=(p2, p98))
    # Equalization
    img = exposure.equalize_hist(img)
    # Adaptive Equalization
    img = exposure.equalize_adapthist(img, clip_limit=0.03)
    return img


def createBinaryAnnotation(img: Union[np.ndarray, torch.Tensor],
                           frg_class: int) -> Union[np.ndarray, torch.Tensor]:
    """
    Creates a binary mask out of the prediction result with root as foreground and the rest as background

    Args:
        img (Union[np.ndarray, torch.Tensor]): Input image
        frg_class (int): value of the foreground class when creating binary segmentation masks

    Raises:
        TypeError: if the input is neither a numpy array or a torch tensor or if the foreground value is wrong. 

    Returns:
        Union[np.ndarray, torch.Tensor]: binary mask 
    """

    if isinstance(img, torch.Tensor):
        u = torch.unique(img)
        bkg = torch.zeros(img.shape)  # background
        if len(u) == 1:
            print("This prediction only contains background")
            return img
        else:
            try: 
                frg = (img == frg_class).int() * 255
            except: 
                print("Error in the foreground value")
    elif isinstance(img, np.ndarray):
        u = np.unique(img)
        bkg = np.zeros(img.shape)  # background
        if len(u) == 1:
            print("This prediction only contains background")
            return img
        else:
            try: 
                frg = (img == frg_class).astype(int) * 255
            except: 
                print("Error in the foreground value")
    else:
        raise TypeError("Input should be a PyTorch tensor or a NumPy array.")
    return bkg + frg

def common_files(folder1, folder2, prefix="Annotation"):
    """
    Compare two folders and return a list of base filenames (ignoring extensions)
    that exist in both folders.
    Files in folder2 may have a prefix (default: 'Annotation') that is ignored.
    """
    def get_basenames(folder, remove_prefix=False):
        basenames = set()
        for f in os.listdir(folder):
            if not os.path.isfile(os.path.join(folder, f)):
                continue
            name, _ = os.path.splitext(f)
            if remove_prefix and name.startswith(prefix):
                name = name[len(prefix):]
            basenames.add(name)
        return basenames

    base1 = get_basenames(folder1)
    base2 = get_basenames(folder2, remove_prefix=True)
    unique = base1.intersection(base2) #intersection()
    return sorted(unique)

    
def get_biomass(binary_img: np.ndarray) -> int:
    """
    Calculate the biomass by counting the number of pixels equal to 1

    Args:
        binary_img (np.ndarray): input image as binary mask

    Returns:
        int: integer value corresponding to the pixel count or root biomass. 
    """
    roi = binary_img > 0
    nerror = 0
    binary_img = binary_img * roi
    biomass = np.unique(binary_img.flatten(), return_counts=True)
    try:
        nbiomass = biomass[1][1]
    except:
        nbiomass = 0
        nerror += 1
        print("Seg error in ")
    return nbiomass




"""
Script for creating small size patches given a specific patch size and sliding window size.
Patches are saved only if at least a certain portion of pixels within the image are annotated to account for class imbalance. 
This percentage is also specified in the configuration file.

Usage:
    pip install rhizonet
    patchify_rhizonet --config_file ./setup_files/setup-prepare.json 
"""

import os
import json
import random
from argparse import ArgumentParser
import argparse
from pathlib import Path

import numpy as np
from skimage import io, util
import os
import sys
import glob
import re
from tqdm import tqdm

from typing import Dict, Tuple, Sequence 

try:
    from .utils import get_image_paths, common_files
except ImportError:
    from utils import get_image_paths, common_files


def parse_prepare_variables(argparse_args):
    """ 
    Parse and merge training variables from a JSON configuration file and command-line arguments.

    Args:
        argparse_args (Namespace): Command-line arguments parsed by argparse.

    Returns:
        Dict: Updated arguments 
    """
    args = vars(argparse_args)
    with open(args["config_file"]) as file_json:
        config_dict = json.load(file_json)
        args.update(config_dict)

    args['train_patch_size'] = tuple(args['train_patch_size'])  
    args['step_size'] = tuple(args['step_size']) 
    return args


def prepare_patches(args: Dict):
    """
    Create small size patches with the patch size and sliding windiw size given in the ``prepare_patches`` configuration file
    Patches are saved only if at least a certain portion of pixels within the image are annotated to account for class imbalance. 
    This percentage is also specified in the configuration file.

    Args:
        args (Dict): dictionary of all processing arguments 

    Returns:
        None: save the image in the specified folder 
    """
    
    image_dir = os.path.join(args['save_path'], "images")
    label_dir = os.path.join(args['save_path'], "labels")
    for folder in [image_dir, label_dir]:
        os.makedirs(folder, exist_ok=True)

    '''For the raw data path, the images and labels paths are given in the setup json file'''
    path_images = get_image_paths(args['images_path'])
    path_labels = get_image_paths(args['labels_path'])
    endpoint_img, endpoint_lbl = Path(path_images[0]).suffix, Path(path_labels[0]).suffix 
    print("Number of images: {} \n Number of labels: {}".format(len(path_images), len(path_labels)))

    # Add condition to keep only matching files in images and labels directory
    label_prefix = args['label_prefix']
    unique_files = common_files(args['images_path'], args['labels_path'], prefix=label_prefix)
    matching_path_images = sorted([os.path.join(args['images_path'], e + endpoint_img) for e in unique_files])
    matching_path_labels = sorted([os.path.join(args['labels_path'], label_prefix + e + endpoint_lbl) for e in unique_files])
    print("Number of maching images: {} \n Number of maching labels: {}".format(len(matching_path_images), len(matching_path_labels)))

    '''
    Note: 
        The following code puts aside m images and labels given that the random split of patches in the training
        builds train/val/test sets that could each contain patches of the same image. 
        That option puts aside `unseen` full-size images.
    '''

    n = len(matching_path_images)
    if args['nb_pred_data'] is not None:
        m = int(args['nb_pred_data'])
        pred_idx = random.sample(list(range(n)), m)
    else:
        pred_idx = []
    train_idx = [e for e in range(n) if e not in pred_idx]

    for i in tqdm(train_idx):
        
        image, label = io.imread(matching_path_images[i]), io.imread(matching_path_labels[i])
        if image.shape[2] >= 3: # RGB or RGBA image
            image_patch_size = args['train_patch_size'] + (image.shape[2], ) 
            image_step_size = args['step_size'] + (3,)
        elif len(image.shape) == 2: # Grayscale image 
            image_patch_size = args['train_patch_size'] 
            image_step_size = args['step_size'] 
        else:
            raise AttributeError("image should be of RGB, RGBA or grayscale")
        
        img_crop = np.vstack(
            util.view_as_windows(image, window_shape=image_patch_size, step=image_step_size))
        img_crop = img_crop.squeeze(1)

        label_crop = np.vstack(
            util.view_as_windows(label, window_shape=args['train_patch_size'], step=args['step_size']))

        for j, (img, lab) in enumerate(zip(img_crop, label_crop)):
            if (np.count_nonzero(lab) / lab.size) > args['min_labeled']:
                f_img = os.path.basename(matching_path_images[i]).split('.')[0]
                f_label = os.path.basename(matching_path_labels[i]).split('.')[0]

                fname_img = f"{f_img}_img_{i:04d}_crop-{j:04d}.tif"
                fname_label = f"{f_label}_img_{i:04d}_crop-{j:04d}.png"

                io.imsave(os.path.join(image_dir, fname_img), util.img_as_ubyte(img), check_contrast=False)
                io.imsave(os.path.join(label_dir, fname_label), util.img_as_ubyte(lab), check_contrast=False)

    if args['nb_pred_data'] is not None:
        pred_image_dir = os.path.join(args['save_pred_path'], 'images')
        pred_label_dir = os.path.join(args['save_pred_path'], 'labels')
        os.makedirs(pred_image_dir, exist_ok=True)
        os.makedirs(pred_label_dir, exist_ok=True)

        for i in tqdm(pred_idx):
            image, label = io.imread(matching_path_images[i]), io.imread(matching_path_labels[i])
            f_img = os.path.basename(matching_path_images[i])
            f_label = os.path.basename(matching_path_labels[i])

            io.imsave(os.path.join(pred_image_dir, f_img), util.img_as_ubyte(image))
            io.imsave(os.path.join(pred_label_dir, f_label), util.img_as_ubyte(label))

    return None


def main():

    parser = ArgumentParser(conflict_handler='resolve', description="Patch cropping parameter setting")
    parser.add_argument("--config_file", type=str,
                        default="../docs/setup_files/setup-prepare.json",
                        help="json file training data parameters")
    args = parser.parse_args()
    args = parse_prepare_variables(args)

    prepare_patches(args)

if __name__ == '__main__':
    main()


import os
import numpy as np
import pytest
from skimage.morphology import disk, dilation, erosion, area_closing, area_opening, convex_hull_image
from skimage.io import imsave, imread
from tempfile import TemporaryDirectory
import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
from rhizonet.postprocessing import getLargestCC, maxProjection, processing


def test_getLargestCC():
    binary_image = np.array([
        [0, 0, 1, 1],
        [0, 0, 1, 1],
        [1, 0, 0, 0],
        [1, 1, 0, 0]
    ])
    
    largest_cc = getLargestCC(binary_image)

    expected_output = np.array([
        [0, 0, 1, 1],
        [0, 0, 1, 1],
        [0, 0, 0, 0],
        [0, 0, 0, 0]
    ], dtype=bool)
    print(largest_cc)

    assert np.array_equal(largest_cc, expected_output), "Largest connected component is incorrect"


def test_maxProjection():
    images = [
        np.array([[0, 1], [2, 3]]),
        np.array([[4, 0], [1, 5]])
    ]

    max_proj = maxProjection(images)
    
    expected_output = np.array([[4, 1], [2, 5]])

    assert np.array_equal(max_proj, expected_output), "Max projection is incorrect"


def test_processing():
    with TemporaryDirectory() as data_dir, TemporaryDirectory() as output_dir:
        # Create dummy input images
        img1 = np.zeros((10, 10), dtype=np.uint8)
        img2 = np.zeros((10, 10), dtype=np.uint8)
        img1[3:7, 3:7] = 1  # Square in the center
        img2[2:8, 2:8] = 1  # Larger square

        os.makedirs(data_dir, exist_ok=True)
        imsave(os.path.join(data_dir, "img1.png"), img1)
        imsave(os.path.join(data_dir, "img2.png"), img2)

        # Run processing
        processing(data_dir, output_dir, area_opening_param=5, area_closing_param=5, disk_radius=2)

        # Validate output
        output_img1 = imread(os.path.join(output_dir, "img1.png"))
        output_img2 = imread(os.path.join(output_dir, "img2.png"))

        # Ensure the convex hull is applied and postprocessing is performed
        assert output_img1.sum() < img1.sum(), "Postprocessing failed for img1"
        assert output_img2.sum() < img2.sum(), "Postprocessing failed for img2"

if __name__ == "__main__":
    pytest.main(["-v", __file__])



import os
import json
import pytest
import torch
import numpy as np
from tempfile import TemporaryDirectory
from skimage import io
import sys
if os.path.exists("/Users/zsordo/Desktop/local.txt"):
    sys.path.append("/Users/zsordo/Desktop/rhizonet")
    
from rhizonet.metrics import calculate_all_metrics, evaluate

# Sample prediction and groundtruth data
@pytest.fixture
def sample_data():
    pred = torch.tensor([
        [1, 1, 0, 0],
        [1, 1, 0, 0],
        [0, 0, 0, 1],
        [0, 0, 1, 1]
    ])
    target = torch.tensor([
        [1, 1, 0, 0],
        [1, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, 1, 1]
    ])
    return pred, target

# Test the metric calculation function
def test_calculate_all_metrics(sample_data):
    pred, target = sample_data
    acc, prec, rec, iou, dice = calculate_all_metrics(pred, target, task="binary", num_classes=2)

    # Expected results
    expected_acc = 15/16 
    expected_prec = 6/7
    expected_rec = 1
    expected_iou = 6/7
    expected_dice = 12/13

    assert acc == pytest.approx(expected_acc, rel=1e-2)
    assert prec == pytest.approx(expected_prec, rel=1e-2)
    assert rec == pytest.approx(expected_rec, rel=1e-2)
    assert iou == pytest.approx(expected_iou, rel=1e-2)
    assert dice == pytest.approx(expected_dice, rel=1e-2)


# Test the evaluation function
def test_evaluate():
    # Create temporary directories for predictions, labels, and logs
    with TemporaryDirectory() as pred_dir, TemporaryDirectory() as label_dir, TemporaryDirectory() as log_dir:
        # Create dummy predictions and labels
        pred1 = torch.tensor([
            [1, 1, 0, 0],
            [1, 1, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 1]
        ], dtype=torch.uint8)*255
        label1 = torch.tensor([
            [1, 1, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 1, 1]
        ], dtype=torch.uint8)*255
        pred_path1 = os.path.join(pred_dir, "pred1.png")
        label_path1 = os.path.join(label_dir, "label1.png")
        io.imsave(pred_path1, pred1)
        io.imsave(label_path1, label1)

        evaluate(pred_dir, label_dir, log_dir, task="binary", num_classes=2)

        # Check that the metrics file is created
        metrics_path = os.path.join(log_dir, "metrics.json")
        assert os.path.exists(metrics_path)

        # Load and verify the metrics
        with open(metrics_path, 'r') as f:
            metrics = json.load(f)
        
        assert "pred1.png" in metrics

        # Expected results
        expected_acc = 15/16 
        expected_prec = 6/7
        expected_rec = 1
        expected_iou = 6/7
        expected_dice = 12/13

        assert metrics["pred1.png"]["accuracy"] == pytest.approx(expected_acc, rel=1e-2)
        assert metrics["pred1.png"]["precision"] == pytest.approx(expected_prec, rel=1e-2)
        assert metrics["pred1.png"]["recall"] == pytest.approx(expected_rec, rel=1e-2)
        assert metrics["pred1.png"]["IOU"] == pytest.approx(expected_iou, rel=1e-2)
        assert metrics["pred1.png"]["Dice"] == pytest.approx(expected_dice, rel=1e-2)


if __name__ == "__main__":
    pytest.main(["-v", __file__])



#!/usr/bin/env python

"""The setup script."""
from os import path
from setuptools import setup, find_packages
import sys
import versioneer

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]

with open(path.join(here, 'requirements-dev.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    dev = [line for line in requirements_file.read().splitlines()
           if not line.startswith('#')]

with open(path.join(here, 'requirements-docs.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    docs = [line for line in requirements_file.read().splitlines()
            if not line.startswith('#')]

setup(
    author="Zineb Sordo, Daniela Ushizima, Peter Andeer, James Sethian and Trent Northen.",
    author_email='zsordo@lbl.gov',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
    description="Segmentation pipeline for EcoFAB images",
    entry_points={
        'console_scripts': [
            'train_rhizonet=rhizonet.train:main',
            'predict_rhizonet=rhizonet.predict:main',
            'postprocess_rhizonet=rhizonet.postprocessing:main',
            'patchify_rhizonet=rhizonet.prepare_patches:main',
            'evalmetrics_rhizonet=rhizonet.metrics:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='rhizonet',
    name='rhizonet',
    packages=find_packages(include=['rhizonet', 'rhizonet.*']),
    test_suite='tests',
    url='https://github.com/lbl-camera/rhizonet',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
    extras_require={
        'tests': ['pytest', 'codecov', 'pytest-cov'],
        'docs': ['sphinx', 'sphinx-rtd-theme', 'myst-parser', 'myst-nb', 'sphinx-panels', 'autodocs']
    },
    setup_requires=["wheel"]
)



import collections
import os

import numpy as np
import matplotlib.pyplot as plt
import pickle

from pytorch_lightning.loggers.logger import Logger
# from pytorch_lightning.loggers import TensorBoardLogger
# from pytorch_lightning.loggers.base import rank_zero_experiment
from pytorch_lightning.utilities import rank_zero_only

import pdb


class mySimpleLogger(Logger):
    def __init__(self, log_dir, keys=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._log_dir = log_dir
        self.keys = ['epoch', 'train_loss_step', 'train_loss_epoch', 'val_loss']
        for k in keys: self.keys.append(k)
        self.history = collections.defaultdict(list) # copy not necessary here
        # The defaultdict in contrast will simply create any items that you try to access

    @property
    def name(self):
        return "my_simple_logger"

    @property
    def experiment(self):
        # Return the experiment version, int or str.
        return "default"

    @property
    def version(self):
        return "1.0"

    @rank_zero_only
    def log_hyperparams(self, params):
        # params is an argparse.Namespace
        # your code to record hyperparameters goes here
        pass

    @rank_zero_only
    def log_metrics(self, metrics, step):
        # metrics is a dictionary of metric names and values
        # your code to record metrics goes here
        for metric_name, metric_value in metrics.items():
            if not metric_name in self.keys:
                continue

            if metric_name != 'epoch':
                self.history[metric_name].append(metric_value)
            else: # case epoch. We want to avoid adding multiple times the same. It happens for multiple losses.
                if (not len(self.history['epoch']) or    # len == 0:
                    not self.history['epoch'][-1] == metric_value) : # the last values of epochs is not the one we are currently trying to add.
                    self.history['epoch'].append(metric_value)
                else:
                    pass
        return

    @rank_zero_only
    def save(self):
        # Optional. Any code necessary to save logger data goes here
        # If you implement this, remember to call `super().save()`
        # at the start of the method (important for aggregation of metrics)
        super().save()
        # with open(f"{self._log_dir}/my_logger_state.pkl", "wb") as f:
        #     pickle.dump(self.state_dict(), f)

    @property
    def log_dir(self):
        return self._log_dir

    @rank_zero_only
    def finalize(self, status):
        # Optional. Any code that needs to be run after training
        # finishes goes here
        plt.figure()
        plt.plot(self.history["train_loss_epoch"], label='train')
        plt.plot(self.history["val_loss"], label='val')
        plt.xlabel('batch steps')
        plt.title("loss")
        plt.gca().legend()
        plt.tight_layout()
        filename = "loss.png"
        plt.savefig(os.path.join(self._log_dir, filename))
        plt.close()

        for key in self.history.keys():
            if not any([x in key for x in ["train_loss_epoch", "val_loss", "epoch"]]):
                plt.figure()
                plt.plot(self.history[key])
                plt.title(key)
                plt.tight_layout()
                filename = os.path.join(self._log_dir, f"{key}.png")
                plt.savefig(filename)
                plt.close()

        for key in self.history.keys():
            arr = np.asarray(self.history[key])
            if arr.ndim == 0:
                arr = np.expand_dims(arr, axis=0)
            filename = os.path.join(self._log_dir, f"{key}.csv")
            np.savetxt(filename, arr, delimiter=',')
